import WebSocket from 'isomorphic-ws';
import { TypedEmitter } from 'tiny-typed-emitter';
import { isClosePeerConnectionMessage, isCommandMessage, isConfigurationMessage, isCreatePeerConnectionMessage, isExperimentStatusChangedMessage, isSignalingMessage, } from './deviceMessages';
import { crosslabTransport, logger } from './logging';
import { WebRTCPeerConnection } from './peer/webrtc-connection';
export class DeviceHandler extends TypedEmitter {
    constructor() {
        super(...arguments);
        this.connections = new Map();
        this.services = new Map();
    }
    async connect(connectOptions) {
        this.ws = new WebSocket(connectOptions.endpoint);
        this.ws.onopen = () => {
            this.ws.send(JSON.stringify({
                messageType: 'authenticate',
                deviceUrl: connectOptions.id,
                token: connectOptions.token,
            }));
            crosslabTransport._set_upstream(info => this.ws.send(JSON.stringify({ messageType: 'logging', content: info })));
        };
        const p = new Promise((resolve, reject) => {
            this.ws.onmessage = authenticationEvent => {
                const authenticationMessage = JSON.parse(authenticationEvent.data);
                if (authenticationMessage.messageType === 'authenticate') {
                    if (authenticationMessage.authenticated) {
                        resolve();
                    }
                    else
                        reject('Authentication failed');
                }
                else {
                    reject(`Expected message with messageType 'authenticate', received ${authenticationMessage.messageType}`);
                }
            };
        });
        this.ws.onclose = event => {
            logger.info('ws closed', { reason: event.reason, code: event.code });
        };
        this.ws.onerror = event => {
            logger.error(event.message, { type: event.type, error: event.error });
        };
        await p;
        this.ws.onmessage = event => {
            const message = JSON.parse(event.data);
            if (isCommandMessage(message)) {
                if (isCreatePeerConnectionMessage(message)) {
                    this.handleCreatePeerConnectionMessage(message);
                }
                else if (isClosePeerConnectionMessage(message)) {
                    this.handleClosePeerConnectionMessage(message);
                }
            }
            if (isSignalingMessage(message)) {
                this.handleSignalingMessage(message);
            }
            if (isConfigurationMessage(message)) {
                this.handleConfigurationMessage(message);
            }
            if (isExperimentStatusChangedMessage(message)) {
                this.handleExperimentStatusChangedMessage(message);
            }
        };
    }
    addService(service) {
        this.services.set(service.serviceId, service);
    }
    handleCreatePeerConnectionMessage(message) {
        if (this.connections.has(message.connectionUrl)) {
            throw Error('Can not create a connection. Connection Id is already present');
        }
        logger.info('creating connection', message);
        const connection = new WebRTCPeerConnection({
            iceServers: [
                { urls: 'stun:stun.goldi-labs.de:3478' },
                { urls: 'turn:turn.goldi-labs.de:3478', username: 'goldi', credential: 'goldi' },
            ],
        });
        connection.tiebreaker = message.tiebreaker;
        this.connections.set(message.connectionUrl, connection);
        for (const serviceConfig of message.services) {
            const service = this.services.get(serviceConfig.serviceId);
            if (service === undefined) {
                throw Error('No Service for the service config was found');
            }
            service.setupConnection(connection, serviceConfig);
        }
        connection.on('signalingMessage', msg => {
            logger.info('sending:', msg);
            this.ws.send(JSON.stringify({
                ...msg,
                messageType: 'signaling',
                connectionUrl: message.connectionUrl,
            }));
        });
        connection.on('connectionChanged', () => {
            const connectionStateChangedMessage = {
                messageType: 'connection-state-changed',
                status: connection.state,
                connectionUrl: message.connectionUrl,
            };
            this.ws.send(JSON.stringify(connectionStateChangedMessage));
            this.emit('connectionsChanged');
        });
        connection.connect();
        this.emit('connectionsChanged');
    }
    handleSignalingMessage(message) {
        const connection = this.connections.get(message.connectionUrl);
        if (connection === undefined) {
            throw Error('No Connection for the signaling message was found');
        }
        connection.handleSignalingMessage(message);
    }
    handleClosePeerConnectionMessage(message) {
        const connection = this.connections.get(message.connectionUrl);
        if (!connection) {
            throw Error('Cannot close a connection. Connection Id is not present');
        }
        logger.info('closing connection', message);
        connection.teardown();
        this.connections.delete(message.connectionUrl);
    }
    handleConfigurationMessage(message) {
        this.emit('configuration', message.configuration);
    }
    handleExperimentStatusChangedMessage(message) {
        this.emit('experimentStatusChanged', {
            status: message.status,
            message: message.message,
        });
    }
    getServiceMeta() {
        return Array.from(this.services).map(service => service[1].getMeta());
    }
}
//# sourceMappingURL=deviceHandler.js.map